/*

 ideserv.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdio.h>
#ifndef WIN32
#include <errno.h>
#ifndef ETHERNET
#ifndef PARPORT
#ifndef RS232
#include <sys/io.h>
#endif
#endif
#endif
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/resource.h>
#else
#include <errno.h>
#include "resource.h"
#include "timeval.h"
#include "sleep.h"
#define lstat(x,y) stat(x,y)
#define S_ISLNK(x) 0
#endif
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>
#ifndef WIN32
#ifdef RS232
#include <sys/ioctl.h>
#include <termios.h>                        //@@@@
#endif
#endif
#include "my_getopt-1.4/getopt.h"
#ifdef ETHERNET
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define sendb(a) do {} while (0)
#endif

FILE *lf=NULL;
unsigned char buffer[16][65536];
unsigned short int buffbytes[16];
unsigned short int buffp[16];
#ifdef ETHERNET
#define SERVER_PORT 64000
int sock;
unsigned char ebuf[1024];
struct sockaddr_in eserver;
struct sockaddr_in eclient;
unsigned int fromlen;
//static int so_reuseaddr=1;
#endif

#include "ideserv.h"
#include "crc8.h"

#define CM_CLOSED 0
#define CM_FILE 1
#define CM_DIR 2
#define CM_ERR 3

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

#ifdef WIN32
/* Prototypes */
INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
long WINAPI PCLinkLoop(long);
#endif

const char *argp_program_version="ideservd 0.22";
FILE *files[16];
int channelmode[16];

void convertfilename(char *s, char typek, char *vessz, char *name, char *type, unsigned char *mode) {
    int b,i,j;

    i=j=*vessz=*mode=0;

    for (b=0;s[b] && s[b]!='/';b++) {
        if (s[b]==':') {i=*vessz=j=*mode=0; continue;}

        switch (*vessz) {
        case 0:
            if (s[b]==typek) {
                (*vessz)++;
                break;
            }
            if (i<16) name[i++]=s[b];
            break;
        case 1:
            if (s[b]==',') {
                (*vessz)++;
                break;
            }
            if (j<3) type[j++]=s[b];
            break;
        case 2:if (mode) *mode=s[b];(*vessz)++;break;
        }
    }
    name[i]=0;
    type[j]=0;
    if (j==1)
        switch (type[0]) {
        case 'S':strcpy(type,"SEQ");break;
        case 'P':strcpy(type,"PRG");break;
        case 'U':strcpy(type,"USR");break;
        case 'B':strcpy(type,"DIR");break;
        case 'J':strcpy(type,"LNK");break;
        }
}

int resolv_path(char *s, char *keszpath, int resolv_all) {
    unsigned char mode;
    char lname[1000],name[17],type[4],vessz;
    struct dirent *ep;
    DIR *dp;
    int found;
    int fel=0,fel2=0;

    while (s[fel] && s[fel]!=':') fel++;
    fel=s[fel]?1:0;

    if (fel) {
        while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') s[fel2++]='/';
        fel=fel2+1;
    }

      if (s[fel]=='/') {
#ifndef WIN32
        keszpath[0]='/';
#else
        keszpath[0]='\\';
#endif

        keszpath[1]=0;
        found=1;
    } else {
        keszpath[0]=0;
        found=0;
    }
    fel=0;

    while (1) {

        while (s[fel] && (s[fel]=='/' || s[fel]==':')) fel++;

        if (!s[fel]) {
            if (!resolv_all) s[0]=0;
            return found;
        }

        if (!resolv_all) {
            fel2=fel;
            while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') fel2++;
            if (!s[fel2]) {
                fel2=fel;fel=0;
                while (s[fel2]) s[fel++]=s[fel2++];
                s[fel]=0;
                return 1;
            }
        }

        dp=opendir(keszpath[0]?keszpath:".");
        if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return 0;}
        convertfilename(s+fel,'/',&vessz,name,type,&mode);

        while ((ep=readdir(dp)))
        {
            struct stat buf;
            DIRECTORY_ENTRY_T dirent;
            char lname2[2020];

            strcpy(lname2,keszpath);
            strncpy(lname,ep->d_name,999);
            strncat(lname2,ep->d_name,999);

            if (lstat(lname2,&buf)) continue;

            createdirentry(ep,&buf,&dirent);

            if ((dirent.attrib & (A_DIR | A_CLOSED))!=(A_CLOSED | A_DIR)) continue;

            dirent.filesize[0]=0;
            if (!matchname(dirent.name,name)) continue;

            found=1;
            break;
        }
        closedir(dp);

        if (!found)
        {
            int f=0,g=fel;
            while (s[g]) {
                lname[f]=c64toascii[(unsigned char)s[g]];
                f++;g++;
            }
            lname[f]=0;
        }

        if ((strlen(lname)+strlen(keszpath)+2)>1000) return 0;

        strcat(keszpath,lname);
#ifndef WIN32
        strcat(keszpath,"/");
#else
        strcat(keszpath,"\\");
#endif
        while (s[fel] && s[fel]!='/' && s[fel]!=':') fel++;

        if (!found) return 0;
    }
}

void do_chdir(char *s) {
    char keszpath[1000];


    if (!resolv_path(s,keszpath,1))
    {
        if (!s[0]) seterror(34,0);
        else seterror(39,0);
        fprintf(lf,"No such directory called \"%s\".\n",keszpath);
        return;
    }
    fprintf(lf,"Chdir to \"%s\".\n",keszpath);
    errtochannel15(chdir(keszpath));
}

void do_mkdir(char *s) {
    unsigned char mode;
    struct dirent *ep;
    DIR *dp;
    char keszpath[1020],vessz,type[4],lname[1000];
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege2;}

    convertfilename(s,',',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;

        found=1;
        break;
    }

    closedir(dp);

    if (!found) {
        int f=0;
        while (s[f]) {
            if (f<16) {
                lname[f]=c64toascii[(unsigned char)s[f]];
                f++;
            }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
//#ifndef WIN32
        errtochannel15(mkdir(keszpath,0777));
//#else
//        errtochannel15(mkdir(keszpath));
//#endif
    }
vege2:
    if (!s[0]) {
        seterror(34,0);
    }
    fprintf(lf,"Mkdir \"%s\".\n",keszpath);
}

void do_scratch(char *s, int smode) {
    unsigned char mode;
    char lname[1000],type[4],vessz,keszpath[1000];
    struct dirent *ep;
    DIR *dp;
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege;}

    convertfilename(s,'=',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
    if (type[0]==0) strcpy(type,"*");

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        }
        else
            if ((dirent.attrib & A_ANY)==A_DIR) continue;

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        found=1;
        break;
    }

    closedir(dp);
    if (!found) seterror(1,0);
vege:
    if (!found) {
        int f=0,g=0;
        while (s[f]) {
            lname[f]=c64toascii[(unsigned char)s[f]];
            f++;
        }
        if (!smode) {
	    if (!relaxfiletypes || memcmp(type,"PRG",3)) {
        	lname[f++]=',';
        	while (type[g]) {
            	    lname[f]=c64toascii[(unsigned char)type[g]];
            	    f++;g++;
        	}
	    }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found) {
        if (smode)
            errtochannel15(rmdir(keszpath));
        else
            errtochannel15(unlink(keszpath));
    }

    if (!s[0]) {
        seterror(34,0);
    }
    if (smode)
        fprintf(lf,"Rmdir \"%s\"\n",keszpath);
    else
        fprintf(lf,"Rm \"%s\"\n",keszpath);
}

void commandchannel(char *s) {
    if (!strncmp(s,"CD:",3) || !strncmp(s,"CD/",3)) {
        s[2]=':';
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (!strncmp(s,"CD_",3)) {
        errtochannel15(chdir(".."));
        fprintf(lf,"Chdir \"..\"\n");
    } else if (!strncmp(s,"C\xd0",2)) {
        seterror(2,1);
        fprintf(lf,"Change to partition %d\n",s[2]);
    } else if (s[0]=='U' && ((s[1] & 0x0f)==9 || (s[1] & 0x0f)==10)) {
        seterror(73,0);
        fprintf(lf,"Identify device\n");
    } else if (!strncmp(s,"I",1)) {
        seterror(0,0);
        fprintf(lf,"Initialize\n");
    } else if (!strncmp(s,"V",1)) {
        seterror(0,0);
        fprintf(lf,"Validate\n");
    } else if (s[0]) {
        seterror(31,0);
    } else fprintf(lf,"\n");
}

void openfile(int usecrc) {
    char s[256];
    unsigned char status=OPEN_ERR;
    unsigned char channel,ch,fp;
    unsigned long long l=0;

#ifdef ETHERNET
    ch=ebuf[1];
#else
    ch=getb();if (tout) return;
#endif
    channel=ch & 0x0f;

    fp=0;
    while (fp<255) {
#ifdef ETHERNET
        s[fp]=ebuf[fp+2];
#else
        s[fp]=getb();if (tout) return;
#endif
        if (!s[fp]) break;
        fp++;
    }
    s[fp]=0;

#ifndef ETHERNET
    if (usecrc) {
        unsigned char crc=crc8[0xCF],gcrc,gt;
        crc=crc8[crc ^ ch];
        fp=0;
        while (s[fp]) { crc=crc8[crc ^ (unsigned char)s[fp]]; fp++;}
	crc=crc8[crc];
        gcrc=getb();if (tout) return;
        crc=crc8[crc ^ gcrc];
        gt=getb();if (tout) return;
        if (crc) {
	    seterror(23,1);
            fprintf(lf,"CRC error %02X: CF %02X ",crc,ch);
            fp=0;
            while (s[fp]) {fprintf(lf,"%02X ",(unsigned char)s[fp]);fp++;}
            fprintf(lf,"00 %02X %02X\n",gcrc,gt);
            return;
        }
    }
#endif //ETHERNET

#ifndef RS232
    alarm(0);
#endif

    if (channel==15) {
	channelmode[15]=CM_ERR;
        fprintf(lf,"Open [%d]: \"%s\": ",channel,s);
	commandchannel(s);
	status=OPEN_STATUS;
    } else {
        char vessz,type[4];
        unsigned char mode;
        
        if (channelmode[channel]==CM_FILE) fclose(files[channel]);
        channelmode[channel]=CM_CLOSED;

        if (s[0]=='$') {
            unsigned char dirline[32];
            struct dirent *ep;
            DIR *dp;
            int sum=0,partition;
            char keszpath[1000];
            channelmode[channel]=CM_DIR;
            buffbytes[channel]=buffp[channel]=0;

	    partition=!strncmp(&s[1],"=P",2);

            resolv_path(s+1,keszpath,0);
            fprintf(lf,"Open [%d]: \"$%s%s\": ",channel,keszpath,s+1);
            convertfilename(s+1,'=',&vessz,s,type,&mode);

            if (s[0]==0) strcpy(s,"*");
            if (type[0]==0) strcpy(type,"*");

            if (channel==1) {
                status=OPEN_ERR;
            } else
            if (channel>1) {
                dp=opendir(keszpath[0]?keszpath:".");
                if (!dp) {
#ifdef ETHERNET
                    if (sendto(sock,"\2\2",2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
                    sendb(2);
#endif
                    errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;
                }
                buffer[channel][buffbytes[channel]++]='I';

                memset(&dirline[0],32,16);
                memset(&dirline[16],0,8);
                dirline[24]=A_DIR;
                memcpy(&dirline[25],"DIR",3);
                memset(&dirline[28],0,4);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,(DIRECTORY_ENTRY_T*)dirline);

                    for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                }

                errtochannel15(closedir(dp));
                status=OPEN_RONLY;
            } else

            if (partition) {
                memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
                memset(&dirline[0x08],32,16);
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"   \42PARTITION 1\42      NET  \0",28);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                errtochannel15(0);
                status=OPEN_RONLY;

            } else {

                dp=opendir(keszpath[0]?keszpath:".");

                if (!dp) {
#ifdef ETHERNET
                    if (sendto(sock,"\2\2",2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
                    sendb(2);
#endif
                    errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
//                sendb(OPEN_RONLY);

                memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
                memset(&dirline[0x08],32,16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    DIRECTORY_ENTRY_T dirent;
                    int offs;
                    unsigned int n;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,&dirent);

                    if (dirent.pointer[0] & 0x80) continue;
                    n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                    if (n>65535) n=65535;

                    dirent.filesize[0]=0;
                    if (!matchname(dirent.name,s)) continue;
                    dirent.time[0]=0;
                    if (!matchname(dirent.filetype,type)) continue;

                    dirline[0x00]=dirline[0x01]=1;
                    dirline[0x02]=n;
                    dirline[0x03]=n >> 8;
                    sum+=n;
                    memset(&dirline[0x04],32,27);
                    offs=7;
                    if (n>9) offs--;
                    if (n>99) offs--;
                    if (n>999) offs--;
                    dirline[offs++]=0x22;
                    n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                    dirline[offs++]=0x22;
                    offs+=(16-n);

                    dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                    n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                    if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                    for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                    if (sum>65535) sum=65535;
                }

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=sum;
                dirline[0x03]=sum >> 8;
                memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                errtochannel15(closedir(dp));
                status=OPEN_RONLY;
            }
        } else {
            char vessz,type[4];
            unsigned char mode,overwrite;
            char lname[1000],keszpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;
            int b;

            overwrite=0;
            for (b=0;s[b];b++) if (s[b]==':') {
                if (s[0]=='@') overwrite=1;
                break;
            }

            resolv_path(s,keszpath,0);
            convertfilename(s,',',&vessz,s,type,&mode);

            dp=opendir(keszpath[0]?keszpath:".");
            if (!dp) {
#ifdef ETHERNET
                if (sendto(sock,"\2\2",2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
                sendb(2);
#endif
                errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
            if (type[0]==0) strcpy(type,(channel==1)?"PRG":"*");

            while ((ep=readdir(dp)))
            {
                struct stat buf;
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strcpy(lname2,keszpath);
                strncpy(lname,ep->d_name,999);
                strncat(lname2,ep->d_name,999);

                if (lstat(lname2,&buf)) continue;

                createdirentry(ep,&buf,&dirent);

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name,s)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;
                break;
            }

            closedir(dp);
            status=OPEN_ERR;

            if (!found) {
                int f=0,g=0;
                while (s[f]) {
                    if (f<16) {
                        lname[f]=c64toascii[(unsigned char)s[f]];
                        f++;
                    }
                }
		if (!relaxfiletypes || memcmp(type,"PRG",3)) {
            	    lname[f++]=',';
            	    while (type[g]) {
                	if (g<3) {
                    	    lname[f]=c64toascii[(unsigned char)type[g]];
                    	    f++;g++;
                	}
            	    }
		}
                lname[f]=0;
            }

            if (channel==0) mode='R';
            if (channel==1) mode='W';

            strcat(keszpath,lname);

            fprintf(lf,"Open [%d]: \"%s\": ",channel,keszpath);

            if (!s[0]) {
                seterror(34,0);goto vege;
            }

            if (mode=='W') {
                int f=0;

                while (lname[f]) {
                    if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                        seterror(33,0);goto vege;
                    }
                    f++;
                }

                if (found && !overwrite) {
                    seterror(63,0);
                } else {
#ifndef WIN32
                    if (overwrite) f=open(keszpath, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
                    else f=open(keszpath, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
#else
                    if (overwrite) f=open(keszpath, O_CREAT|O_WRONLY|O_TRUNC|O_BINARY, S_IRUSR|S_IWUSR);
                    else f=open(keszpath, O_CREAT|O_EXCL|O_WRONLY|O_BINARY, S_IRUSR|S_IWUSR);
#endif
                    if (f<0) errtochannel15(1); else {
                        status=OPEN_WONLY;
                        files[channel]=fdopen(f,"wb");
                        channelmode[channel]=CM_FILE;
                        errtochannel15(0);
                    }
                }
            } else {
                if (!found) {
                    seterror(62,0);
                } else {
                    files[channel]=fopen(keszpath,"rb");
                    if (files[channel]) {
                        status=OPEN_RONLY;
                        channelmode[channel]=CM_FILE;
                        errtochannel15(0);
                    } else errtochannel15(1);
                }
            }
        }
    }
vege:
#ifdef ETHERNET
    ebuf[0]=status;ebuf[1]=status;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
    sendb(status);
#endif
    
}

void closefile(int usecrc) {
    unsigned char channel,ch;
#ifdef ETHERNET
    channel=(ch=ebuf[1]) & 0x0f;
    ebuf[0]=0;ebuf[1]=0;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
    channel=(ch=getb()) & 0x0f;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xC3];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
    sendb(0);
#endif
    fprintf(lf,"Close [%d]\n",channel);
    if (channelmode[channel]==CM_FILE) fclose(files[channel]);
    channelmode[channel]=CM_CLOSED;
}

void readfile(int usecrc) {
    unsigned char channel,ch,status=0;
    unsigned short int bytes,i,j;
    struct timeval start,end;
    unsigned long long l=0,l2;

#ifdef ETHERNET
    channel=(ch=ebuf[1]) & 0x0f;
    bytes=ebuf[2];
    if (!bytes) bytes=256;
#else
    channel=(ch=getb()) & 0x0f;if (tout) return;
    bytes=getb();if (tout) return;
    bytes|=getb()*256;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xD2];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ (bytes & 0xff)];
        crc=crc8[crc ^ ((bytes >> 8) & 0xff)];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
#endif
    fprintf(lf,"Read [%d] %d bytes\n",channel,bytes);
#ifndef ETHERNET
    sendb(0);if (tout) return;
#endif
    if (channelmode[channel]==CM_CLOSED) {
#ifdef ETHERNET
        ebuf[0]=0;ebuf[1]=0;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
	sendb(0);if (tout) return;
	sendb(0);if (tout) return;
	if (usecrc) {
	    unsigned char crc=crc8[crc8[0]];
	    sendb(crc);if (tout) return;
	}
        sendb(0);
#endif
        return;
    }
    if (channelmode[channel]==CM_ERR) {
	j=buffp[channel];
	if (bytes==(buffbytes[channel]-j)) status=0x40;
	if (bytes>buffbytes[channel]-j) {bytes=buffbytes[channel]-j;status=0x42;}
	l=bytes;
	buffp[channel]+=bytes;
    } else {
	if (channelmode[channel]==CM_FILE) {
	    unsigned long itt;
	    l=bytes=fread(buffer[channel],1,bytes?bytes:65536,files[channel]);
	    if (feof(files[channel])) status=0x42; else
	    {
		itt=ftell(files[channel]);
		fgetc(files[channel]);
		if (feof(files[channel])) status=0x40;
		fseek(files[channel],itt,SEEK_SET);
	    }
	    j=0;
	} else {
	    j=buffp[channel];
	    if (bytes==(buffbytes[channel]-j)) status=0x40;
	    if (bytes>buffbytes[channel]-j) {bytes=buffbytes[channel]-j;status=0x42;}
	    l=bytes;
	    buffp[channel]+=bytes;
	}
    }
#ifdef ETHERNET
    ebuf[0]=bytes>0;
    ebuf[1]=bytes;
#else //ETHERNET
    sendb(~bytes);if (tout) return;
    sendb(~bytes >> 8);if (tout) return;
 #ifdef RS232
    alarm(3);
  #ifndef WIN32
    tcdrain(commhandle);
  #else //WIN32
    FlushFileBuffers(commhandle);
  #endif //WIN32
    alarm(0);if (tout) return;
 #endif //RS232
#endif //ETHERNET
#ifdef ETHERNET
    ebuf[(bytes & 0xfffe)+3]=status;
    ebuf[((bytes+1) & 0xfffe)+2]=status;
    ebuf[((bytes+1) & 0xfffe)+3]=status;
    for (i=0;i<bytes;i++,j++) ebuf[1+bytes-i]=buffer[channel][j];
    if (sendto(sock,ebuf,((bytes+1) & 0xfffe)+4, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
    gettimeofday(&start,NULL);
    if (bytes) {
        unsigned short int j2=j;
        for (i=0;i<bytes;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	if (usecrc) {
	    unsigned char crc=0;
	    crc=crc8[crc ^ (unsigned char)(~bytes)];
	    crc=crc8[crc ^ (unsigned char)(~bytes >> 8)];
	    for (i=0;i<bytes;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
//    fprintf(lf,"crc: %02X\n", crc);
	    sendb(crc);if (tout) return;
	}
    }else {
	unsigned short int j2=j;
	for (i=0;i<32768;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	for (i=0;i<32768;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	if (usecrc) {
	    unsigned char crc=0;
	    for (i=0;i<32768;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
	    for (i=0;i<32768;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
	    sendb(crc);if (tout) return;
	}
    }
    gettimeofday(&end,NULL);
    l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
    if (l2) {
	l=l*1000000/l2;
	fprintf(lf,"Read %d bytes at %lld bytes/seconds\n",bytes,l);
    }
    sendb(status);//fprintf(lf,"Status: %02X\n",status);
#endif // ETHERNET
#ifdef RS232
    alarm(1);
#ifndef WIN32
    tcdrain(commhandle);
#else
    FlushFileBuffers(commhandle);
#endif
    alarm(0);if (tout) return;
#endif
    if (channel==15 && buffbytes[15]==buffp[15]) seterror(0,0);
}


void writefile(int usecrc) {
    unsigned char channel,ch;
    unsigned short int bytes,i;
    struct timeval start,end;
    unsigned long long l=0,l2;

#ifdef ETHERNET
    channel=(ch=ebuf[1]) & 0x0f;
    bytes=ebuf[2];
#else
    channel=(ch=getb()) & 0x0f;if (tout) return;
    bytes=getb();if (tout) return;
    bytes|=getb()*256;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xD7];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ (bytes & 0xff)];
        crc=crc8[crc ^ ((bytes >> 8) & 0xff)];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
#endif //ETHERNET
    fprintf(lf,"Write [%d] %d bytes\n",channel,bytes);
    if (channelmode[channel]==CM_ERR) {
        char s[256];
        unsigned char j;
#ifdef ETHERNET
        ebuf[0]=0;
        j=0;
        for (i=0;i<bytes;i++) {s[j]=ebuf[i+4];if (j<255) j++;}
#else //ETHERNET
        sendb(0);if (tout) return;
        j=0;
        for (i=0;i<bytes;i++) {s[j]=getb();if (tout) return;if (j<255) j++;}

	if (usecrc) {
    	    unsigned char crc=0;

	    for (i=0;i<j;i++) crc=crc8[crc ^ (unsigned char)s[i]];
    	    crc=crc8[crc ^ getb()];if (tout) return;
	    getb();if (tout) return;
    	    if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
	} else {
    	    getb();if (tout) return;
	}
#endif //ETHERNET

        s[j]=0;
        if (j && s[j-1]==0x0d) s[j-1]=0;

#ifdef ETHERNET
        ebuf[1]=0;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
        sendb(0);
#endif
        commandchannel(s);
    } else {
        if (channelmode[channel]!=CM_FILE) {fprintf(lf,"Not writeable.\n");sendb(2);return;}
#ifdef ETHERNET
        if (fwrite(ebuf+4,1,bytes,files[channel])==bytes) ebuf[0]=ebuf[1]=0; else ebuf[0]=ebuf[1]=2;
        if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) exit(1);
#else
        sendb(0);if (tout) return;

        gettimeofday(&start,NULL);
        for (i=0;i<bytes;i++) {buffer[channel][i]=getb();if (tout) return;}
        gettimeofday(&end,NULL);
	
	if (usecrc) {
    	    unsigned char crc=0;

	    for (i=0;i<bytes;i++) crc=crc8[crc ^ buffer[channel][i]];
    	    crc=crc8[crc ^ getb()];if (tout) return;
	    getb();if (tout) return;
    	    if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
	} else {
    	    getb();if (tout) return;
	}
        if (fwrite(buffer[channel],1,bytes,files[channel])==bytes) sendb(0); else sendb(2);
        l=bytes;
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
	if (l2) {
	    l=l*1000000/l2;
    	    fprintf(lf,"Wrote %d bytes at %lld bytes/seconds\n",bytes,l);
	}
#endif
    }
}

struct arguments_t
{
    int background;
    int priority;
    int relaxfiletypes;
    char *user;
    char *group;
    char *root;
    char *log;
#ifdef PARPORT
    char *parport;
#else
#ifndef RS232
    unsigned short int lptport;
#endif
#endif
#ifdef RS232
    char *serport;
#endif
#ifdef ETHERNET
    struct in_addr sin_addr;
    unsigned char network;
#endif
    int hog;
};

struct arguments_t arguments;

#ifndef WIN32
void testarg(int argc,char *argv[]) {
    int c;

    for (;;) {
        static struct option long_options[]= {
            {"user", required_argument, 0, 'u'},
            {"group", required_argument, 0, 'g'},
            {"root", required_argument, 0, 'r'},
            {"log", required_argument, 0, 'l'},
            {"nice", required_argument, 0, 'n'},
            {"allprg", no_argument, 0, 'P'},
            {"help", no_argument, 0, '?'},
            {"usage", no_argument, 0, 1},
            {"version", no_argument, 0, 'V'},
#ifndef RS232
            {"hog", no_argument, 0, 'h'},
#ifndef PARPORT
            {"lptport", required_argument, 0, 'p'},
#endif
#endif
            {"background", no_argument, 0, 'b'},
#ifdef PARPORT
            {"device", required_argument, 0, 'd'},
#endif
#ifdef RS232
            {"device", required_argument, 0, 'd'},
#endif
#ifdef ETHERNET
            {"ipaddress", required_argument, 0, 'i'},
            {"network", required_argument, 0, 'N'},
#endif
            {0,0,0,0}
        };
        int option_index=0;

        c=getopt_long(argc, argv,
                      "u:g:r:l:n:P?V"
#ifndef RS232
                      "h"
#ifndef PARPORT
#ifndef ETHERNET
                      "p:"
#endif
#endif
#endif
                      "b"
#ifdef PARPORT
                      "d:"
#endif
#ifdef RS232
                      "d:"
#endif
#ifdef ETHERNET
                      "i:N:"
#endif
                      , long_options, &option_index);
        if (c==-1) {
            if (optind==argc)
                break;
            c='?';
        }

        switch (c) {
        case '?':
            printf(
                   "Usage: ideservd [OPTION...]\n"
                   "IDEDOS 0.9x PCLink fileserver\n"
                   "\n"
                   "  -b, --background\t     Fork into background\n"
#ifdef PARPORT
                   "  -d, --device=PARPORT\t     Parport device (/dev/parport0)\n"
#endif
#ifdef RS232
                   "  -d, --device=SERIALPORT    Serial port device (/dev/ttyS0)\n"
#endif
                   "  -g, --group=GROUP\t     Group under we run (nogroup)\n"
#ifndef RS232
                   "  -h, --hog\t\t     CPU hog (off)\n"
#endif
#ifdef ETHERNET
                   "  -i, --ipaddress=IP\t     IP address of C64 on network\n"
#endif
                   "  -l, --log=FILE\t     Logfile (stdout)\n"
                   "  -n, --nice=ADJUST\t     Adjust nice level (-19)\n"
#ifdef ETHERNET
                   "  -N, --network=NUM\t     Network number (0)\n"
#endif
#ifndef RS232
#ifndef PARPORT
#ifndef ETHERNET
                   "  -p, --lptport=IOPORT\t     Printer port address (0x378)\n"
#endif
#endif
#endif
                   "  -P, --allprg\t\t     Almost everything to PRG\n"
                   "  -r, --root=DIRECTORY\t     Root directory (.)\n"
                   "  -u, --user=USER\t     User under we run (nobody)\n"
                   "  -?, --help\t\t     Give this help list\n"
                   "      --usage\t\t     Give a short usage message\n"
                   "  -V, --version\t\t     Print program version\n"
                   "\n"
                   "Mandatory or optional arguments to long options are also mandatory or optional\n"
                   "for any corresponding short options.\n"
                   "\n"
                   "Report bugs to <soci@c64.rulez.org>.\n"
                   );
                   exit(0);
        case 1:
            printf(
                   "Usage: ideservd [-bhP?V] [-g GROUP] [-l FILE] [-n ADJUST] [-p IOPORT]\n"
                   "            [-r DIRECTORY] [-u USER] [--background] [--group=GROUP] [--hog]\n"
                   "            [--log=FILE] [--nice=ADJUST] [--lptport=IOPORT] [--allprg]\n"
                   "            [--root=DIRECTORY] [--user=USER] [--help] [--usage] [--version]\n"
                  );
            exit(0);
        case 'V':
            printf("%s\n",argp_program_version);
            exit(0);
        case 'b':arguments.background=1;break;
        case 'h':arguments.hog=1;break;
        case 'P':arguments.relaxfiletypes=1;break;
        case 'u':arguments.user=optarg;break;
        case 'g':arguments.group=optarg;break;
        case 'l':arguments.log=optarg;break;
        case 'r':arguments.root=optarg;break;
#ifdef PARPORT
        case 'd':arguments.parport=optarg;break;
#else
#ifndef RS232
#ifndef ETHERNET
        case 'p':arguments.lptport=strtol(optarg,NULL,0);break;
#endif
#endif
#endif
#ifdef RS232
        case 'd':arguments.serport=optarg;break;
#endif
#ifdef ETHERNET
        case 'i':
            {
                struct hostent *hp;
                hp = gethostbyname(optarg);
                if (hp) memcpy(&arguments.sin_addr, hp->h_addr, sizeof(arguments.sin_addr));
            }
        case 'N':arguments.network=strtol(optarg,NULL,0) & 0xff;break;
#endif
        case 'n':arguments.priority=strtol(optarg,NULL,0);break;
        default:
            exit(1);
        }
    }
}
#else // WIN32
void testarg(int argc,char *argv[]) {
    int j;
    for (j=1;j<argc;j++)
    {
	if (!strcmp(argv[j],"-?") || !strcmp(argv[j],"--help")) {
                MessageBox (NULL,
		"Usage: ideservd [OPTION...]\n"
		"IDEDOS 0.9x PCLink fileserver\n"
                "\n"
#ifdef RS232
                "  -d SERIALPORT\t     Serial port device (COM1)\n"
#else
                "  -h, --hog\t     CPU hog (off)\n"
#ifdef ETHERNET
                "  -i, --ipaddress\t      IP address of C64 on network\n"
                "  -N, --network\t      Network number (0)\n"
#endif
                "  -p, --lptport\t     Printer port address (0x378)\n"
#endif
                "  -P, --allprg\t     Almost everything to PRG\n"
                "  -l FILE\t\t     Logfile (ideservd.log)\n"
//                "  -n, --nice=ADJUST\t     Adjust nice level (-19)\n"
		"  -?, --help\t     Give this help list\n"
		"      --usage\t     Give a short usage message\n"
		"  -V, --version\t     Print program version\n"
		"\n"
//		"Mandatory or optional arguments to long options are also mandatory or optional\n"
//		"for any corresponding short options.\n"
		"\n"
		"Report bugs to <soci@c64.rulez.org>.\n"
		,"Ideservd usage",MB_OK);
	    exit(1);
	}
        if (!strcmp(argv[j],"--usage")) {
#ifdef RS232
            MessageBox(NULL,"Usage: ideservd [-dhpP?V] [-d SERIALPORT] [-l FILE] [--allprg] [--hog] [--help]\n"
                   "\t    [--usage] [--version]\n" ,"Ideservd usage",MB_OK);
#else
#ifdef ETHERNET
            MessageBox(NULL,"Usage: ideservd [-hpP?V] [-l FILE] [-i IP] [--allprg] [--hog] [--help] [--usage] [--version]\n" ,"Ideservd usage",MB_OK);
#else
            MessageBox(NULL,"Usage: ideservd [-hpP?V] [-l FILE] [-p IOPORT] [--allprg] [--hog] [--help] [--usage] [--version]\n" ,"Ideservd usage",MB_OK);
#endif
#endif
	    exit(1);
	}
	if (!strcmp(argv[j],"-V") || !strcmp(argv[j],"--version")) {
	    MessageBox(NULL,argp_program_version,"Ideservd version",MB_OK);
	    exit(1);
	}
	if (!strcmp(argv[j],"-P") || !strcmp(argv[j],"--allprg")) {arguments.relaxfiletypes=1;continue;}
#ifndef RS232
	if (!strcmp(argv[j],"-h") || !strcmp(argv[j],"--hog")) {arguments.hog=1;continue;}
	if (!strcmp(argv[j],"-p") || !strcmp(argv[j],"--lptport")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.lptport=strtol(argv[j],NULL,0);
	    continue;
	}
#endif
	if (!strcmp(argv[j],"-l")) {
	    j++;if (j>=argc) goto ide2;
//	    if (arguments.log) goto ide3;
	    arguments.log=malloc(strlen(argv[j])+1);
	    strcpy(arguments.log,argv[j]);
	    continue;
        }
#ifdef RS232
	if (!strcmp(argv[j],"-d")) {
	    j++;if (j>=argc) goto ide2;
//	    if (arguments.serport) goto ide3;
	    arguments.serport=malloc(strlen(argv[j])+1);
	    strcpy(arguments.serport,argv[j]);
	    continue;
        }
#endif
#ifdef ETHERNET
	if (!strcmp(argv[j],"-i") || !strcmp(argv[j],"--ipaddress")) {
            struct hostent *hp;
	    j++;if (j>=argc) goto ide2;
            hp = gethostbyname(argv[j]);
            if (hp) memcpy(&arguments.sin_addr, hp->h_addr, sizeof(arguments.sin_addr));
	    continue;
	}
	if (!strcmp(argv[j],"-N") || !strcmp(argv[j],"--network")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.network=strtol(argv[j],NULL,0) & 0xff;
	    continue;
	}
#endif

/*        if (!strcmp(argv[j],"-n")) {
	    int i=0;
	    j++;if (j>=argc)*/if (0) {
		ide2:
		MessageBox(NULL,"Ideservd: option requires an argument!",NULL,MB_ICONERROR|MB_OK);
		goto ide;
	    }
/*
	    while (argv[j][i] && argv[j][i]!='=') i++;
            if (argv[j][i]=='=') {
                argv[j][i]=0;
                arguments.priority=strtol(argv[j],NULL,0);
            }
            continue;
        }*/
	ide3:
	MessageBox(NULL,"Usage: ideservd [OPTION...]\n",NULL,MB_OK);
	ide:
        MessageBox(NULL,"Try `ideservd --help' or `ideservd --usage' for more information.\n",NULL,MB_OK);
        if (arguments.log) free(arguments.log);
#ifdef RS232
        if (arguments.serport) free(arguments.serport);
#endif
        exit(1);
    }
}
#endif

int vtime=1;

void init(int argc,char *argv[])
{
    struct passwd *user2;
    struct group *group2;
    arguments.log=NULL;
#ifdef ETHERNET
    arguments.sin_addr.s_addr=htonl(INADDR_ANY);
#endif
#ifdef RS232
    arguments.serport=NULL;
#endif
    arguments.background=0;
    arguments.user= "nobody";
    arguments.group= "nogroup";
    arguments.root= ".";
#ifdef PARPORT
    arguments.parport= "/dev/parport0";
#else
 #ifndef RS232
    arguments.lptport=DEFAULT_IOPORT;
 #endif
#endif
    arguments.priority=-19;
    arguments.hog=0;
    arguments.relaxfiletypes=0;
    testarg(argc, argv);

#ifdef RS232
    if (arguments.serport==NULL)
 #ifdef WIN32
        arguments.serport= "COM1";
 #else
        arguments.serport= "/dev/ttyS0";
 #endif
#endif
    if (arguments.log==NULL)
#ifdef WIN32
        arguments.log="ideservd.log";
#else
        arguments.log="/dev/stdout";
#endif
        relaxfiletypes=arguments.relaxfiletypes;
#ifndef ETHERNET
 #ifdef RS232
    initport(arguments.serport);
 #else
  #ifndef PARPORT
    IOPORT=arguments.lptport;
  #endif
  #ifdef WIN32
   #ifndef WIN32_IO
     /* Load the library */
     hLib = LoadLibrary("inpout32.dll");

     if (hLib == NULL) {fprintf(stderr,"LoadLibrary Failed.\n");exit(1);}

     /* get the address of the function */

     inp32 = (inpfuncPtr) GetProcAddress(hLib, "Inp32");

     if (inp32 == NULL) {fprintf(stderr,"GetProcAddress for Inp32 Failed.\n");exit(1);}

     oup32 = (oupfuncPtr) GetProcAddress(hLib, "Out32");

     if (oup32 == NULL) {fprintf(stderr,"GetProcAddress for Oup32 Failed.\n");exit(1);}
   #endif // !WIN32_IO
  #else

   #ifdef PARPORT
     parportfd=open(arguments.parport, O_RDWR);
     if (parportfd == -1) {fprintf(stderr,"Opening parport device\n");exit(1);}
     if (ioctl(parportfd, PPCLAIM)) {
         close(parportfd);
         fprintf(stderr,"Cannot access parport device\n");exit(1);
     }
     {
         int mode=IEEE1284_MODE_BYTE;
         if (ioctl(parportfd, PPSETMODE, &mode)) {
             close(parportfd);
             fprintf(stderr,"Cannot set byte mode on parport device\n");exit(1);
         }

         mode=0;
         if (ioctl(parportfd, PPDATADIR, &mode)) {
             close(parportfd);
             fprintf(stderr,"Cannot set direction on parport device\n");exit(1);
         }
     }
   #else
    if (ioperm(IOPORT,3,1)) {fprintf(stderr,"Cannot allocate IO ports\n");exit(1);}
   #endif // PARPORT
  #endif // WIN32 (LPT port)
 #endif // RS232
#else //ETHERNET
     sock=socket(AF_INET, SOCK_DGRAM, 0);
     if (sock < 0) {fprintf(stderr,"Cannot open socket\n");exit(1);}
     memset(&eserver,0,sizeof(eserver));
     eserver.sin_family=AF_INET;
     eserver.sin_addr.s_addr=SERVER_IP;
     eserver.sin_port=htons(SERVER_PORT+arguments.network);
//     setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &so_reuseaddr, sizeof(so_reuseaddr));
     if (bind(sock,(struct sockaddr *)&eserver,sizeof(eserver))<0) {fprintf(stderr,"Cannot bind\n");exit(1);}
     fromlen = sizeof(struct sockaddr_in);
#endif

#ifndef WIN32
    nice(arguments.priority);
     if (!(user2=getpwnam(arguments.user))) {fprintf(stderr,"User %s does not exists.\n",arguments.user);exit(1);}
     if (!(group2=getgrnam(arguments.group))) {fprintf(stderr,"Group %s does not exists.\n",arguments.group);exit(1);}
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
#endif
     if (!(lf=fopen(arguments.log,"wt"))) {fprintf(stderr,"Cannot open logfile '%s'\n",arguments.log);exit(1);}
#ifndef WIN32
     if (chroot(arguments.root) || chdir("/")) {fprintf(stderr,"Changing root dir to '%s'\n",arguments.root);exit(1);}
     if (setgid(group2->gr_gid) || setuid(user2->pw_uid)) {fprintf(stderr,"Could not drop privileges.\n");exit(1);}
     if (arguments.background) if (fork()) {fprintf(stderr,"Forking into background.\n");exit(0);}
     fclose(stdin);
     fclose(stderr);
     fclose(stdout);
#endif
    if (arguments.hog) vtime=0;
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch(message) 
  {
   case WM_INITDIALOG:
        return TRUE;

   case WM_COMMAND:
      switch(wParam) 
      {
         case IDC_EXIT:
	 PostQuitMessage(0);
         EndDialog(hwnd, TRUE);
         return TRUE;
      }
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
// int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    unsigned char b;
    int dyntime=1,n;
    init(argc,argv);
    fprintf(lf,"%s\n",argp_program_version);
    signal(SIGALRM,timeout);
    signal(SIGTERM,terminate);
    signal(SIGINT,terminate);
    for (b=0;b<16;b++) channelmode[b]=CM_CLOSED;
    seterror(73,0);

#ifdef WIN32
    HWND hwnd;
    MSG msg;
    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);	// probably not necessary...
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,NULL,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    terminate();
}

long WINAPI PCLinkLoop(long lParam)
{
    unsigned char b;
    int dyntime=1,n;
#endif //WIN32

start:
    fprintf(lf,"ready.\n");
    dyntime=1;
start2:
#ifndef ETHERNET
 #ifndef RS232
    alarm(0);
  #ifdef PC64
   #ifdef PARPORT
    for (;;) {
	unsigned char b;
	ioctl(parportfd, PPRSTATUS, &b);
	if (b>=0x80) {
	    ioctl(parportfd, PPRSTATUS, &b);
	    if (b & 0x78) break;
	}
        if (vtime) {
            usleep(dyntime);
            if (dyntime<16384) dyntime<<=1;
        }
    }
   #else
    do {
        while (inb(STATPORT)<0x80) if (vtime) {
            usleep(dyntime);
            if (dyntime<16384) dyntime<<=1;
        }
    } while ((inb(STATPORT) & 0x78)==0);
   #endif // PARPORT
    tout=0;alarm(1);
  #else // PC64
    out_b(CLKLO+DATAHI+ATNHI);
    wdatalo() if (vtime) {
        usleep(dyntime);
        if (dyntime<16384) dyntime<<=1;
    }
    out_b(CLKHI+DATAHI+ATNHI);
    tout=0;alarm(1);
    wdatahi() {
        if (vtime) {
            usleep(dyntime);
            if (dyntime<16384) dyntime<<=1;
        }
        if (tout) goto start;
    } //set clklo, wait atnlo
  #endif // PC64
 #else // RS232
    tout=2;
 #endif // RS232
    b=getb();
#ifdef RS232
    if (tout) goto start2;
#else
    if (tout) goto start;
    alarm(0);
#endif
#else // ETHERNET
    n = recvfrom(sock,ebuf,sizeof(ebuf),0,(struct sockaddr *)&eclient,&fromlen);
    eclient.sin_addr.s_addr=arguments.sin_addr.s_addr;
    if (n < 0) exit(1);
    ebuf[n]=0;
    b=ebuf[0];

#endif // ETHERNET
    switch (b)
    {
//    case 'X':return 0;
    case 0x00:goto start2;
    case 'O':openfile(0);break;
    case 0xCF:openfile(1);break;
    case 'R':readfile(0);break;
    case 0xD2:readfile(1);break;
    case 'W':writefile(0);break;
    case 0xD7:writefile(1);break;
    case 'C':closefile(0);break;
    case 0xC3:closefile(1);break;
    default:fprintf(lf,"Unknown command %02X\n",b);
    }
#ifdef RS232
    alarm(1);tout=0;
#ifndef WIN32
    tcdrain(commhandle);
#else
    FlushFileBuffers(commhandle);
#endif
    alarm(0);
#endif
    goto start;
}
